// ***********************************************************************
// **                                                                   **
// ** OpenCM 9.04 Firmware for EZ-Builder (c)Synthiam Inc.              **
// **                                                                   **
// ** Setup & information at:                                           **
// ** https://synthiam.com/Hardware/Robotis-OpenCM-9-04-17530           **
// **                                                                   **
// ** Updated: April 8, 2019                                            **
// **                                                                   **
// ***********************************************************************

#include <DynamixelSDK.h>
#include <Servo.h>

// The first digital port that is usable on this controller
#define _PortStart 0

// The last digital port on this controller that is usable
#define _PortCnt 26

// The number of analog ports
#define _AnalogPorts 10

// The firmware version that is reported to EZ-Builder to notify of capabilities
#define _FIRMWARE_ID 0x00000007

// The communication baud rate
#define _BAUD_RATE 921600

// The primary communication interface between EZ-Builder and this controller
// If you wish to use the Bluetooth module, change this to the desired port (i.e. Serial2)
// If you're using bluetooth, you will also need to change the _BAUD_RATE to 9600 or what ever
// the bluetooth module uses
#define COMMUNICATION_PORT Serial

// The amount of RX buffer on the communication interface for EZ-Builder
#define _BUFFER_SIZE 1024

// The amount of RX buffer on the Serial 1 (dynamixel ttl) port
// This is because the default input buffer on the Serial1 is only
// 60 bytes or something... so this buffer will store RX data until
// it's queried by EZ-Builder.
#define _BUFFER_SIZE1 1024

// --------------------------------------------------------------------------

// EZB COMM BUFFER
byte         _INPUT_BUFFER[_BUFFER_SIZE];
unsigned int _WRITE_POSITION = 0;
unsigned int _READ_POSITION = 0;

// Serial1 (dynamixel ttl) COM BUFFER
byte         _INPUT_BUFFER1[_BUFFER_SIZE1];
unsigned int _WRITE_POSITION1 = 0;
unsigned int _READ_POSITION1 = 0;
bool         _SERIAL_1_INITIALIZED = false;

Servo Servos[_PortCnt];

#define CmdUnknown            0
#define CmdReleaseAllServos   1
#define CmdGetUniqueID        2
#define CmdEZBv3              3
#define CmdEZBv4              4
#define CmdSoundBeep          5
#define CmdEZServo            6
#define CmdI2CWrite           10
#define CmdI2CRead            11
#define CmdBootLoader         14
#define CmdSetPWMSpeed        15
#define CmdSetServoSpeed      39
#define CmdPing               0x55
#define CmdSetDigitalPortOn   100
#define CmdSetDigitalPortOff  124
#define CmdGetDigitalPort     148
#define CmdSetServoPosition   172
#define CmdGetADCValue        196
#define CmdSendSerial         204
#define CmdHC_SR04            228
#define CmdGetFirwareID       253
#define CmdSoundStreamCmd     254

// CmdEZBv4 Commands
// ----------------------------------------------------------------------------------
#define CmdV4SetLipoBatteryProtectionState 0
#define CmdV4SetBatteryMonitorVoltage      1
#define CmdV4GetBatteryVoltage             2
#define CmdV4GetCPUTemp                    3

#define CmdV4UARTExpansion0Init            5
#define CmdV4UARTExpansion0Write           6
#define CmdV4UARTExpansion0AvailableBytes  7
#define CmdV4UARTExpansion0Read            8

#define CmdV4UARTExpansion1Init            9
#define CmdV4UARTExpansion1Write           10
#define CmdV4UARTExpansion1AvailableBytes  11
#define CmdV4UARTExpansion1Read            12

#define CmdV4UARTExpansion2Init            13
#define CmdV4UARTExpansion2Write           14
#define CmdV4UARTExpansion2AvailableBytes  15
#define CmdV4UARTExpansion2Read            16

#define CmdV4I2CClockSpeed                 17
#define CmdV4UARTClockSpeed                18
#define CmdV4ResetToDefaults               19

// CmdSoundStreamCmd Commands
// ----------------------------------------------------------------------------------
#define CmdSoundInitStop 0
#define CmdSoundLoad     1
#define CmdSoundPlay     2

// For EZB Comm
bool IsAvail() {

  return _WRITE_POSITION != _READ_POSITION || COMMUNICATION_PORT.available();
}

// For Serial1
bool IsAvail1() {

  return _WRITE_POSITION1 != _READ_POSITION1 || Serial1.available();
}

// For EZB Com
int Avail() {

  if (_READ_POSITION > _WRITE_POSITION)
    return (_BUFFER_SIZE - _READ_POSITION) + _WRITE_POSITION + COMMUNICATION_PORT.available();

  return (_WRITE_POSITION - _READ_POSITION) + COMMUNICATION_PORT.available();
}

// For Serial1
int Avail1() {

  if (_READ_POSITION1 > _WRITE_POSITION1)
    return (_BUFFER_SIZE1 - _READ_POSITION1) + _WRITE_POSITION1 + Serial1.available();

  return (_WRITE_POSITION1 - _READ_POSITION1) + Serial1.available();
}

// For EZB Com
int32_t Read32() {

  return ReadByte() | (ReadByte() << 8) | (ReadByte() << 16) | (ReadByte() << 24);
}

// For Serial1
int32_t Read321() {

  return ReadByte1() | (ReadByte1() << 8) | (ReadByte1() << 16) | (ReadByte1() << 24);
}

// For EZB Com
int16_t Read16() {

  return ReadByte() | (ReadByte() << 8);
}

// For Serial1
int16_t Read161() {

  return ReadByte1() | (ReadByte1() << 8);
}

// For EZB Com
byte ReadByte() {

  while (_WRITE_POSITION == _READ_POSITION && COMMUNICATION_PORT.available() == 0);

  while (COMMUNICATION_PORT.available()) {

    _WRITE_POSITION++;

    _INPUT_BUFFER[_WRITE_POSITION % _BUFFER_SIZE] = COMMUNICATION_PORT.read();
  }

  _READ_POSITION++;

  return _INPUT_BUFFER[_READ_POSITION % _BUFFER_SIZE];
}

// For Serial1
byte ReadByte1() {

  while (_WRITE_POSITION1 == _READ_POSITION1 && Serial1.available() == 0);

  while (Serial1.available()) {

    _WRITE_POSITION1++;

    _INPUT_BUFFER1[_WRITE_POSITION1 % _BUFFER_SIZE1] = Serial1.read();
  }

  _READ_POSITION1++;

  return _INPUT_BUFFER1[_READ_POSITION1 % _BUFFER_SIZE1];
}

// For Serial1
void UpdateBuffer1() {

  if (!_SERIAL_1_INITIALIZED)
    return;

  while (Serial1.available()) {

    _WRITE_POSITION1++;

    _INPUT_BUFFER1[_WRITE_POSITION1 % _BUFFER_SIZE1] = Serial1.read();
  }
}

void setup() {

  COMMUNICATION_PORT.begin(_BAUD_RATE);

  // Half duplex control pin for Serial1
  // Arduino pin assignments: https://github.com/ROBOTIS-GIT/OpenCM9.04/blob/master/arduino/opencm_arduino/opencm9.04/variants/OpenCM904/chip.h
  // Schematic http://support.robotis.com/en/baggage_files/opencm/opencm904_rev_10_final_schematic.pdf
  // High is for writing, Low is for reading
  // We will initialize LOW so data is read to ring buffer in the background interrupt
  // and will change to HIGH only during writes of Serial1
  pinMode(28, OUTPUT);
  digitalWrite(28, LOW);
}

void loop() {

  UpdateBuffer1();

  doEZProtocol();
}

void Write32(long val) {

  COMMUNICATION_PORT.write((byte)(val & 0xff));
  COMMUNICATION_PORT.write((byte)((val >> 8) & 0xff));
  COMMUNICATION_PORT.write((byte)((val >> 16) & 0xff));
  COMMUNICATION_PORT.write((byte)((val >> 24) & 0xff));
}

void Write16(int val) {

  COMMUNICATION_PORT.write((byte)(val & 0xff));
  COMMUNICATION_PORT.write((byte)((val >> 8) & 0xff));
}

void doEZProtocol() {

  if (!IsAvail())
    return;

  byte cmd = ReadByte();

  if (cmd == CmdPing) {

    // return as a "Capability Controller"
    COMMUNICATION_PORT.write(222);
  } else if (cmd == CmdGetFirwareID) {

    Write32(_FIRMWARE_ID);
  } else if (cmd == CmdReleaseAllServos) {

    for (int port = _PortStart; port < _PortCnt; port++)
      if (Servos[port].attached())
        Servos[port].detach();
  } else if (cmd >= CmdSetServoPosition && cmd <= CmdSetServoPosition + 23) {

    byte port = cmd - CmdSetServoPosition;
    byte pos = ReadByte();

    if (port >= _PortStart && port <= _PortCnt) {

      if (pos == 0 && Servos[port].attached()) {

        Servos[port].detach();
      } else {

        if (!Servos[port].attached())
          Servos[port].attach(port);

        Servos[port].write(pos);
      }
    }
  } else if (cmd >= CmdSetPWMSpeed && cmd <= CmdSetPWMSpeed + 23) {

    byte port = cmd - CmdSetPWMSpeed;
    byte pos = ReadByte();

    if (port >= _PortStart && port <= _PortCnt) {

      if (Servos[port].attached())
        Servos[port].detach();

      pinMode(port, OUTPUT);

      analogWrite(port, map(pos, 0, 100, 0, 255));
    }
  } else if (cmd >= CmdSetDigitalPortOn && cmd <= CmdSetDigitalPortOn + 23) {

    byte port = cmd - CmdSetDigitalPortOn;

    if (port >= _PortStart && port <= _PortCnt) {

      if (Servos[port].attached())
        Servos[port].detach();

      pinMode(port, OUTPUT);

      digitalWrite(port, HIGH);
    }
  } else if (cmd >= CmdSetDigitalPortOff && cmd <= CmdSetDigitalPortOff + 23) {

    byte port = cmd - CmdSetDigitalPortOff;

    if (port >= _PortStart && port <= _PortCnt) {

      if (Servos[port].attached())
        Servos[port].detach();

      pinMode(port, OUTPUT);

      digitalWrite(port, LOW);
    }
  } else if (cmd >= CmdGetDigitalPort && cmd <= CmdGetDigitalPort + 23) {

    byte port = cmd - CmdGetDigitalPort;

    if (port >= _PortStart && port <= _PortCnt) {

      if (Servos[port].attached())
        Servos[port].detach();

      pinMode(port, INPUT);

      COMMUNICATION_PORT.write(digitalRead(port));
    } else {

      COMMUNICATION_PORT.write((byte)0);
    }
  } else if (cmd >= CmdGetADCValue && cmd <= CmdGetADCValue + 7) {

    byte port = cmd - CmdGetADCValue;

    if (port >= 0 && port <= _AnalogPorts)
      Write16(analogRead(port));
    else
      Write16(0);
  } else if (cmd == CmdEZBv4) {

    byte cmd2 = ReadByte();

    // ------------ UART #0 (i.e. Serial1 only for Half Duplex Communication)
    if (cmd2 == CmdV4UARTExpansion0Init) {

      // Very undocumented command that enables the UART 1 (Serial1) to B6/B7 (29/30) from A9/A10 (11/12)
      // Half duplex communication which requires pin 28 to specific direction
      Serial1.setDxlMode(true);

      Serial1.begin(Read32());

      _SERIAL_1_INITIALIZED = true;
    } else if (cmd2 == CmdV4UARTExpansion0Write) {

      // Half duplex buffer in TX mode
      digitalWrite(28, HIGH);

      // Delay for the duplex buffer to switch over
      delay(10);

      uint16_t length = Read16();

      for (uint16_t x = 0; x < length; x++)
        Serial1.write(ReadByte());

      // Flush the output send buffer before we change duplex buffer direction
      Serial1.flush();

      // Half duplex buffer in RX mode
      digitalWrite(28, LOW);
    } else if (cmd2 == CmdV4UARTExpansion0AvailableBytes) {

      Write16(Avail1());
    } else if (cmd2 == CmdV4UARTExpansion0Read) {

      uint16_t length = Read16();

      for (uint16_t x = 0; x < length; x++)
        COMMUNICATION_PORT.write(ReadByte1());
    }

    // ------------ UART #1 (i.e. Serial2)
    if (cmd2 == CmdV4UARTExpansion1Init) {

      Serial2.begin(Read32());
    } else if (cmd2 == CmdV4UARTExpansion1Write) {

      uint16_t length = Read16();

      for (uint16_t x = 0; x < length; x++)
        Serial2.write(ReadByte());
    } else if (cmd2 == CmdV4UARTExpansion1AvailableBytes) {

      Write16(Serial2.available());
    } else if (cmd2 == CmdV4UARTExpansion1Read) {

      uint16_t length = Read16();

      for (uint16_t x = 0; x < length; x++)
        COMMUNICATION_PORT.write(Serial2.read());
    }

    // ------------ UART #2 (i.e. Serial3)
    if (cmd2 == CmdV4UARTExpansion2Init) {

      Serial3.begin(Read32());
    } else if (cmd2 == CmdV4UARTExpansion2Write) {

      uint16_t length = Read16();

      for (uint16_t x = 0; x < length; x++)
        Serial3.write(ReadByte());
    } else if (cmd2 == CmdV4UARTExpansion2AvailableBytes) {

      Write16(Serial3.available());
    } else if (cmd2 == CmdV4UARTExpansion2Read) {

      uint16_t length = Read16();

      for (uint16_t x = 0; x < length; x++)
        COMMUNICATION_PORT.write(Serial3.read());
    }
  }
}
